@@ -1,6 +1,6 @@
 ###############################################################################
-# Copyright 2017 - Climate Research Division
-#                  Environment and Climate Change Canada
+# Copyright 2017-2021 - Climate Research Division
+#                       Environment and Climate Change Canada
 #
 # This file is part of the "fstd2nc" package.
 #
@@ -41,17 +41,18 @@ class FSTD_Compat (BufferBase):
     """
 
     # Check if compatibility interface should be activated.
-    fstd_compat = kwargs.pop('fstd_compat', False)
-    if fstd_compat:
-      self._fstluk = self._fstluk_compat
+    self._fstd_compat = kwargs.pop('fstd_compat', False)
+    if self._fstd_compat:
       self.to_netcdf = self._to_netcdf_compat
     super(FSTD_Compat,self).__init__(*args,**kwargs)
 
   # Override fstluk to keep track of which FST records were used in the
   # conversion.  The header information for these records will be added to the
   # output file.
-  def _fstluk_compat (self, rec_id, dtype=None, rank=None, dataArray=None):
+  def _fstluk (self, rec_id, dtype=None, rank=None, dataArray=None):
     import numpy as np
+    if not self._fstd_compat:
+      return super(FSTD_Compat,self)._fstluk(rec_id, dtype, rank, dataArray)
     if not hasattr(self,'_used_rec_ids'):
       self._used_rec_ids = []
     # If rec_id is a dict from rpnpy, then convert it to an index.
@@ -181,7 +182,10 @@ class FSTD_Compat (BufferBase):
       if hasattr(var,'array'):
         # Tell netCDF4 to use big-endian encoding of data, where applicable.
         # FSTD requires big-endian encoding (and netCDF4 can work with either).
-        v = f.createVariable(var.name, datatype=var.array.dtype.newbyteorder('>'), endian='big', dimensions=var.dims, zlib=zlib, complevel=compression)
+        if var.array.dtype.itemsize > 1:
+          v = f.createVariable(var.name, datatype=var.array.dtype.newbyteorder('>'), endian='big', dimensions=var.dims, zlib=zlib, complevel=compression)
+        else:
+          v = f.createVariable(var.name, datatype=var.array.dtype, dimensions=var.dims, zlib=zlib, complevel=compression)
         # Write the metadata.
         v.setncatts(var.atts)
         direct_addresses[var.array.tobytes()] = write_data (v, (), var.array)
@@ -251,15 +255,8 @@ class FSTD_Compat (BufferBase):
     # Include all metadata records.  Difficult to know for certain which ones
     # were used, since librmn may have read them internally in some routines
     # like horizontal / vertical grid extraction.
-    nomvars = np.array(self._headers['nomvar'])
-    ni = np.array(self._headers['ni'])
-    nj = np.array(self._headers['nj'])
-    for nomvar in self._meta_records:
-      nomvar = (nomvar+'    ')[:4]
-      used_rec_ids.update(np.where(nomvars==nomvar)[0])
-    for nomvar in self._maybe_meta_records:
-      nomvar = (nomvar+'    ')[:4]
-      used_rec_ids.update(np.where((nomvars==nomvar)&((ni==1)|(nj==1)))[0])
+    ismeta = self._headers['ismeta']
+    used_rec_ids.update(np.where(ismeta==1)[0])
 
     # Prepare the file for writing FSTD structures.
     with open(filename,'r+b') as f:
