@@ -151,9 +151,18 @@ class ExternOutput (BufferBase):
             # Special case: already have a dask object from external source.
             # (E.g., from fstpy)
             if hasattr(self, '_extern_table'):
-              nodes = list(self._extern_table['d'].iloc[rec_id].dask.values())
-            if hasattr(self, '_extern_table') and len(nodes) == 0:
-              dsk[key] = nodes[0]
+              d = self._extern_table['d'].iloc[rec_id]
+              if hasattr(d,'compute'):
+                # Start a dask graph using the external dask array as the source.
+                dsk[key] = (d.compute,)
+                # Ensure the dask array includes the degenerate outer dimensions.
+                # Otherwise get a runtime error if slicing is done on this.
+                dsk[key] = (np.ravel, dsk[key], 'K')
+                dsk[key] = (np.reshape, dsk[key], chunk_shape)
+              else:  # Special case: have a numpy array in memory.
+                dsk[key] = d
+                dsk[key] = np.ravel(dsk[key], 'K')
+                dsk[key] = np.reshape(dsk[key], chunk_shape)
             # Otherwise, construct one with our own dask wrapper.
             else:
               dsk[key] = (_preferred_chunk_order,filename,rec_key,(self._read_chunk, rec_id, chunk_shape, var.dtype))
@@ -240,11 +249,18 @@ class ExternOutput (BufferBase):
     import numpy as np
     from fstpy.dataframe import add_grid_column
     from fstpy.std_io import add_dask_column
+    # Special case: our data is already from an fstpy table, not from an FSTD
+    # file in our control.
+    # E.g., if some smartass does Buffer.from_fstpy(df).to_fstpy()
+    if hasattr(self, '_extern_table'):
+      return self._extern_table
     # Put all the header info into a dictionary.
     fields = ['nomvar', 'typvar', 'etiket', 'ni', 'nj', 'nk', 'dateo', 'ip1', 'ip2', 'ip3', 'deet', 'npas', 'datyp', 'nbits', 'grtyp', 'ig1', 'ig2', 'ig3', 'ig4', 'datev']
     table = dict()
+    # Create a mask to exclude deleted / overwritten records.
+    mask = self._headers['dltf'] == 0
     for field in fields:
-      col = self._headers[field]
+      col = self._headers[field][mask]
       # Convert byte arrays to strings, which is what fstpy expects.
       if col.dtype.str.startswith('|S'):
         col = np.asarray(col,dtype=col.dtype.str.replace('|S','<U'))
@@ -256,8 +272,8 @@ class ExternOutput (BufferBase):
     # Temporarily insert some extra columns needed for the data.
     table['shape'] = list(zip(table['ni'],table['nj']))
     filenames = dict((i,f) for i,f in enumerate(self._files))
-    table['path'] = pd.Series(self._headers['file_id']).map(filenames)
-    table['key'] = (self._headers['key'] << 10)
+    table['path'] = pd.Series(self._headers['file_id'][mask]).map(filenames)
+    table['key'] = (self._headers['key'][mask] << 10)
     # Generate dask objects
     #TODO: use our own, in case we modified the data?
     # (doesn't normally happen, but you never know...)
