@@ -61,6 +61,7 @@ class netCDF_Atts (BufferBase):
     except ImportError:
       import configparser as ConfigParser
     from collections import OrderedDict
+    import numpy as np
     metadata_file = kwargs.pop('metadata_file',None)
     if metadata_file is None:
       metafiles = []
@@ -92,7 +93,7 @@ class netCDF_Atts (BufferBase):
         if k == 'units': continue
         try:
           metadata[varname][k] = float(v) # First, try converting to float.
-          metadata[varname][k] = int(v) # Try further conversion to int.
+          metadata[varname][k] = np.int32(v) # Try further conversion to int.
         except ValueError: pass
     self._metadata = metadata
     # Check for renames.
@@ -127,7 +128,11 @@ class netCDF_Atts (BufferBase):
     for obj in self._iter_objects():
       # Add extra metadata provided by the user?
       if obj.name in self._metadata:
-        obj.atts.update(self._metadata[obj.name])
+        for attname, attval in self._metadata[obj.name].items():
+          # Check if attribute should be removed?
+          # (i.e., if value is intentionally left empty)
+          if attval == "": obj.atts.pop(attname,None)
+          else: obj.atts[attname] = attval
       # Apply renames.
       if obj.name in renames:
         obj.name = renames[obj.name]
@@ -265,7 +270,12 @@ class netCDF_IO (BufferBase):
     def get_var_id (var):
       out = []
       for fmt in self._human_var_id:
-        out.append(fmt%var.atts)
+        try:
+          out.append(fmt%var.atts)
+        except KeyError:
+          out.append(None)
+      # Check if no useful information found to construct any part of the id.
+      if list(set(out)) == [None]: raise KeyError
       return tuple(out)
 
     # Generate unique variable names.
@@ -294,7 +304,12 @@ class netCDF_IO (BufferBase):
       # maintaining uniqueness.
       for j in reversed(range(len(var_ids))):
         test = var_ids[:j] + var_ids[j+1:]
-        if len(set(zip(*test))) == len(var_list):
+        # If variables are still unique without this part of the id, then it
+        # can be safely removed.
+        # Also, can remove this part if it's not consistently defined.
+        # (E.g. if one of the versions of the variable does not have a
+        # consistent version of an attribute to use).
+        if len(set(zip(*test))) == len(var_list) or None in var_ids[j]:
           var_ids = test
 
       var_ids = zip(*var_ids)
@@ -309,7 +324,7 @@ class netCDF_IO (BufferBase):
 
     for var in self._iter_objects():
       # Names must start with a letter or underscore.
-      if not var.name[0].isalpha():
+      if not var.name[0].isalpha() and var.name[0] != '_':
         warn(_("Renaming '%s' to '_%s'.")%(var.name,var.name))
         var.name = '_'+var.name
 
