@@ -63,8 +63,10 @@ def gdgaxes (gdid):
 # Base class for grid mapping classes
 class GridMap(object):
   # Mean radius of the Earth used in OGC CRS (coord. ref. system) descriptions
-  # Note that Earth's mean radius = 6371000. m in librmn!
-  _earth_radius = 6371229. 
+  # This is the value used in librmn, and also corresponds to the Normal
+  # Sphere approximation referenced in
+  # https://proj.org/usage/ellipsoids.html#built-in-ellipsoid-definitions
+  _earth_radius = 6370997.
   def __init__(self, grd):
 #   grd is a dictionary returned by readGrid containing grid parameters 
     from collections import OrderedDict
@@ -279,10 +281,14 @@ class PolarStereo(GridMap):
     self._ay = ( np.rint(xy['y'][0,:]) - 1) * self._res
     # Determine the false easting and northing from 
     # the coordinates of the pole and of grid point (1,1)
-    _pole_pi = self._grd['pi']
-    _pole_pj = self._grd['pj']
-    self._false_easting =  self._ax[int(round(_pole_pi)) - 1] - self._ax[0]
-    self._false_northing = self._ay[int(round(_pole_pj)) - 1] - self._ay[0]
+    if self._grd['north']:
+      pole = gdxyfll (self._grd['id'], 90, 0)
+    else:
+      pole = gdxyfll (self._grd['id'], -90, 0)
+    px = np.rint(pole['x'][0] - 1) * self._res
+    py = np.rint(pole['y'][0] - 1) * self._res
+    self._false_easting =  px - self._ax[0]
+    self._false_northing = py - self._ay[0]
     self._xaxisatts['long_name'] = 'x-coordinate of polar-stereographic projection'
     self._xaxisatts['standard_name'] = 'projection_x_coordinate'
     self._xaxisatts['units'] = 'm'
@@ -358,6 +364,9 @@ class XYCoords (BufferBase):
     # Special case for series data - match any of the lat/lon grids.
     if var.atts['grtyp'] in ('+','Y'):
       header = fstlir (self._meta_funit, nomvar=coordname, rank=3)
+      # Make sure this actually matches a grid of the correct shape.
+      if header['ni'] != var.atts['ni'] or header['nj'] != var.atts['nj']:
+        header = None
     else:
       header = fstlir (self._meta_funit, nomvar=coordname, ip1=var.atts['ig1'],
                              ip2=var.atts['ig2'], ip3=var.atts['ig3'], rank=3)
@@ -496,7 +505,7 @@ class XYCoords (BufferBase):
           # Calculate the mean lat/lon arrays in double precision.
           meanlat = np.mean(np.array(latarray,dtype=float),axis=1,keepdims=True)
           meanlon = np.mean(np.array(lonarray,dtype=float),axis=0,keepdims=True)
-          if latarray.shape[1] > 1 and lonarray.shape[1] > 1 and np.allclose(latarray,meanlat) and np.allclose(lonarray,meanlon):
+          if latarray.shape[0] > 1 and lonarray.shape[1] > 1 and np.allclose(latarray,meanlat) and np.allclose(lonarray,meanlon):
             # Reduce back to single precision for writing out.
             meanlat = np.array(meanlat,dtype=latarray.dtype).squeeze()
             meanlon = np.array(meanlon,dtype=lonarray.dtype).squeeze()
@@ -512,9 +521,9 @@ class XYCoords (BufferBase):
             gridaxes = [lat,lon]
 
           # Case 2: lat/lon are series of points.
-          elif latarray.shape[0] == 1 and lonarray.shape[0] == 1 and ('i' in var.dims or 'station_id' in var.dims):
-            latarray = latarray[0,:]
-            lonarray = lonarray[0,:]
+          elif 1 in latarray.shape and 1 in lonarray.shape and ('i' in var.dims or 'station_id' in var.dims):
+            latarray = latarray.squeeze()
+            lonarray = lonarray.squeeze()
             # Special case for station data
             station_id = var.getaxis('station_id')
             if station_id is not None:
@@ -604,6 +613,18 @@ class XYCoords (BufferBase):
         if var.name == 'LO' and ('lon' in var.dims or lon in coordinates):
           var.name = None
 
+    # Only create distinct grid_mappings when they're actually distinct.
+    # (may have duplicates because of different ni,nj from staggered grid.)
+    gridmaps = dict()
+    for var in varlist:
+      if 'grid_mapping' in var.atts:
+        gmapvar = var.atts['grid_mapping']
+        key = tuple(sorted(gmapvar.atts.items()))
+        if key not in gridmaps:
+          gridmaps[key] = gmapvar
+        gmapvar = gridmaps[key]
+        var.atts['grid_mapping'] = gmapvar
+
     self._varlist = [v for v in varlist if v.name is not None]
 
     # Special case - there are no data records, ONLY a pair of lat/lon
